(define-module (alterator vm option)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-13)
  :use-module (alterator vm defs)
  :use-module (alterator vm utils)
  :export (task-options))

(define (probe-option option)
  (format #t "PROBE  ---------------------------------------------------------\n~S\n" option)
  option)

(define (set-task-option! name value)
  (and-let*
   ((task (fluid-ref current-task)))
   (set-procedure-property!
    task
    'context
    (assv-set! (procedure-property task 'context)
               name value))))

(define (task-option name)
  (and-let*
   ((task (fluid-ref current-task))
    (options (procedure-property task 'context)))
   (assv-ref options name)))

(define (option-importance< opta optb)
  (and
   (or (memq 'important (assv-ref opta 'flags))
       (memq 'not-required (assv-ref optb 'flags))) #t))

(define (filter-option:custom option)
  (and option
       (case (fluid-ref current-operation)
         ((op-pt-create)
          ;; no OS/2 selection, we're linux
          #f)
         ((op-segment-create)
          ;; TypeByName should be enough
          (cond ((equal? (assv-ref option 'name) "TypeByNumber")
                 (set-task-option! 'type-by-number (assv-ref option 'value)) #f)
                ((equal? (assv-ref option 'name) "Bootable") #f)
                (#t option)))
         ((op-volume-mkfs)
          (cond ((equal? (assv-ref option 'name) "version") #f)
                (#t option)))
         (else option))))

(define (filter-option:active option)
  (and option
       (if (memq 'inactive (assv-ref option 'flags)) #f
           option)))

(define (fourth-primary? seg-or-disk)
  (let* ((disk (if (eqv? 'segment (obj->type seg-or-disk))
                   (car (obj->children seg-or-disk))
                   seg-or-disk))
         (segs (obj->parents disk)))
    (and (< (length (filter segment-meta? segs)) 2)
         (< 2 (length (filter segment-data? segs))))))

(define (mangle-option:custom option)
  (cond
   ((member (assv-ref option 'name) '("Size" "Offset"))
    (assv-set! option 'flags `(important . ,(assv-ref option 'flags))))
   ((equal? (assv-ref option 'name) "Primary")
    (cond ((and (assv-ref option 'value)
                (not (task-option 'force-extended))
                (fourth-primary? (fluid-ref current-object)))
           (set! ((fluid-ref current-task) 'options) '(("Primary" . #f)))
           (set-task-option! 'force-extended #t)
           (assv-set! option 'value #f))
          ((not (assv-ref option 'value))
           (assv-set! option 'flags `(inactive . ,(assv-ref option 'flags))))
          (#t option)))
   ((equal? (assv-ref option 'name) "TypeByName")
    (let ((option
           (assv-set! option 'constraint
                      (reverse (cons "Linux RAID"
                                     (cdr (reverse (delete "HPFS" (assv-ref option 'constraint)))))))))
      (cond ((equal? (task-option 'type-by-number) "253")
             (assv-set!
              (assv-set! option 'value "Linux RAID")
              'flags `(lasthop . ,(assv-ref option 'flags))))
            ((equal? (assv-ref option 'value) "Linux LVM")
             (assv-set! option 'flags `(lasthop . ,(assv-ref option 'flags))))
            (#t option))))
   (#t option)))

(define (mangle-option:to-value option)
  (acons 'to-value
         (case (assv-ref option 'widget)
           ((slider)
            (let ((step (caddr (assv-ref option 'constraint))))
              (lambda (v) (inexact->exact (* step v)))))
           ((combobox combobox/ex edit)
            (case (assv-ref option 'type)
              ((number) string->number)                           ; #f on ""
              ((string) (lambda (s) (if (string-null? s) #f s)))  ; #f on ""
              (else values)))
           (else values))
         option))

(define (scale-generic min max step)
  (values
   #f
   (lambda(v)
     (number->string (* step v)))
   (lambda(s)
     (and-let*
      ((v (false-if-exception (inexact->exact (/ (string->number s) step)))))
      (and (<= min v max) v)))))

(define (scale-megabytes min max step)
  (values
   "MB"
   (lambda(v)
     (number->string (inexact->exact (* step v))))
   (lambda(s)
     (and-let*
      ((v (false-if-exception (inexact->exact (/ (string->number s) step)))))
      (and (<= min v max) v)))))

(define (scale-sectors-staged min max step)
  ;; assume bytes/sector eqv 512, vg reports all-zero geom
  (let* ((gbs (< 20971520 (* step (- max min))))  ; ~10G
         (units (if gbs "GB" "MB"))
         (scale (if gbs (/ step 2097152) (/ step 2048)))
         (rscale (if gbs (/ 2097152 step) (/ 2048 step))))
    (values
     units
     (lambda(v)
       (number->string (inexact->exact (* v scale))))
     (lambda(s)
       (and-let*
        ((v (false-if-exception (inexact->exact (* (string->number s) rscale)))))
        (and (<= min v max) v))))))

(define (scale-sectors min max step)
  ;; assume bytes/sector eqv 512, vg reports all-zero geom
  (let* ((scale (/ step 2048))
         (rscale (/ 2048 step)))
    (values
     "MB"
     (lambda(v)
       (number->string (inexact->exact (* v scale))))
     (lambda(s)
       (and-let*
        ((v (false-if-exception (inexact->exact (* (string->number s) rscale)))))
        (and (<= min v max) v))))))

(define (mangle-option:scale option)
  (if (eq? 'slider (assv-ref option 'widget))
      (let ((constraint (assv-ref option 'constraint)))
        (cond ((apply = 0 constraint)
               (acons 'scale (values "MB" (lambda(v) "0") (lambda(s) 0)) option))
              ((eqv? 'sector (assv-ref option 'unit))
               (acons 'scale (apply scale-sectors constraint) option))
              ((eqv? 'mb (assv-ref option 'unit))
               (acons 'scale (apply scale-megabytes constraint) option))
              (#t (acons 'scale (apply scale-generic constraint) option))))
      option))

(define (mangle-option:value option)
  (let ((value (assv-ref option 'value)))
    (case (assv-ref option 'widget)
      ((edit)
       (assv-set! option 'value
                  (if (number? value)
                      (number->string value)
                      (or value ""))))
      ((slider)
       (assv-set! option 'value
                  (let ((vstep (caddr (assv-ref option 'constraint))))
                    (if (= 0 vstep) value
                        (inexact->exact (/ value vstep))))))
      (else option))))

(define (mangle-option:pixmap option)
  (acons 'pixmap
         (if (memq 'not-required (assv-ref option 'flags))
             ""
             "reqfield.gif")
         option))

(define (mangle-option:widget option)
  (cond
   ((eq? 'boolean (assv-ref option 'type))
    (acons 'widget 'checkbox option))
   ((eq? 'list (assv-ref option 'constraint-type))
    (acons 'widget
           (if (memq 'value-is-list (assv-ref option 'flags))
               'checklist
               (if (memq 'expandable (assv-ref option 'flags))
                   'combobox/ex
                   'combobox))
           option))
   ((eq? 'range (assv-ref option 'constraint-type))
    ;; convert range to list, when appropriate
    ;; or normalize min/max vs step
    (let ((vmin (car (assv-ref option 'constraint)))
          (vmax (cadr (assv-ref option 'constraint)))
          (vstep (caddr (assv-ref option 'constraint))))
      (acons 'widget 'slider
             (assv-set! option 'constraint
                        (list (if (= vstep 0) vmin (inexact->exact (/ vmin vstep)))
                              (if (= vstep 0) vmax (inexact->exact (/ vmax vstep)))
                              vstep)))))
   (#t (acons 'widget 'edit option))))

(define (mangle-option:flags option)
  (let ((flags (assv-ref option 'flags)))
    (acons 'ready (or (and (memq 'not-required flags) #t)
                      (not (memq 'no-initial-value flags)))
           option)))

(define (mangle-option:constraint option)
  (let ((constraint (assv-ref option 'constraint)))
    (acons 'type (car constraint)
           (acons 'constraint-type (cadr constraint)
                  (acons 'constraint (caddr constraint)
                         (alist-delete 'constraint option))))))

(define (mangle-option:name option)
  (acons 'name (car option) (cdr option)))

(define (mangle-option option)
  (filter-option:custom
   (filter-option:active
    (mangle-option:custom
     (mangle-option:to-value
      (mangle-option:scale
       (mangle-option:value
        (mangle-option:pixmap
         (mangle-option:widget
          (mangle-option:flags
           (mangle-option:constraint
            (mangle-option:name option))))))))))))

(define (task-options op obj task)
  (with-fluids
   ((current-operation op)
    (current-object obj)
    (current-task task))
   (sort
    (filter-map mangle-option (task 'options))
    option-importance<)))
